import cpuinfo
import psutil
from rich import box
from rich.padding import Padding
from rich.panel import Panel
from rich.table import Table
from textual.widget import Widget

from .braille_stream import BrailleStream


def val_to_color(val: float, minval: float, maxval: float) -> str:
    t = (val - minval) / (maxval - minval)
    k = round(t * 3)
    return {0: "color(4)", 1: "color(6)", 2: "color(6)", 3: "color(2)"}[k]


# https://stackoverflow.com/a/312464/353337
def chunks(lst, n):
    for i in range(0, len(lst), n):
        yield lst[i : i + n]


# https://stackoverflow.com/a/6473724/353337
def transpose(lst):
    return list(map(list, zip(*lst)))


def flatten(lst):
    return [item for sublist in lst for item in sublist]


class CPU(Widget):
    def on_mount(self):
        self.is_first_render = True
        self.width = 0
        self.height = 0

        # self.max_graph_width = 200

        num_cores = psutil.cpu_count(logical=False)
        num_threads = psutil.cpu_count(logical=True)

        # 8 threads, 4 cores -> [0, 4, 1, 5, 2, 6, 3, 7]
        assert num_threads % num_cores == 0
        self.core_order = flatten(
            transpose(list(chunks(range(num_threads), num_cores)))
        )

        self.cpu_total_stream = BrailleStream(50, 7, 0.0, 100.0)

        self.cpu_percent_streams = [
            BrailleStream(10, 1, 0.0, 100.0)
            for _ in range(num_threads)
            # BlockCharStream(10, 1, 0.0, 100.0) for _ in range(num_threads)
        ]

        try:
            temps = psutil.sensors_temperatures()
        except AttributeError:
            self.has_temps = False
        else:
            self.has_temps = "coretemp" in temps
            if self.has_temps:
                temp_low = 20.0
                temp_high = temps["coretemp"][0].high
                assert temp_high is not None
                self.temp_total_stream = BrailleStream(
                    50, 7, temp_low, temp_high, flipud=True
                )
                self.core_temp_streams = [
                    BrailleStream(5, 1, temp_low, temp_high) for _ in range(num_cores)
                ]

        self.box_title = ", ".join(
            [
                f"{num_threads} thread" + ("s" if num_threads > 1 else ""),
                f"{num_cores} core" + ("s" if num_cores > 1 else ""),
            ]
        )

        self.brand_raw = cpuinfo.get_cpu_info()["brand_raw"]
        self.set_interval(2.0, self.collect_data)

    def collect_data(self):
        # CPU loads
        self.cpu_total_stream.add_value(psutil.cpu_percent())
        #
        load_indiv = psutil.cpu_percent(percpu=True)
        cpu_percent_colors = [val_to_color(val, 0.0, 100.0) for val in load_indiv]
        for stream, load in zip(self.cpu_percent_streams, load_indiv):
            stream.add_value(load)

        # CPU temperatures
        if self.has_temps:
            coretemps = psutil.sensors_temperatures()["coretemp"]
            self.temp_total_stream.add_value(coretemps[0].current)
            #
            for stream, temp in zip(self.core_temp_streams, coretemps[1:]):
                stream.add_value(temp.current)

        lines_cpu = self.cpu_total_stream.graph
        last_val_string = f"{self.cpu_total_stream.values[-1]:5.1f}%"
        lines0 = lines_cpu[0][: -len(last_val_string)] + last_val_string
        lines_cpu = [lines0] + lines_cpu[1:]
        #
        cpu_total_graph = "[color(4)]" + "\n".join(lines_cpu) + "[/]\n"

        #
        if self.has_temps:
            lines_temp = self.temp_total_stream.graph
            last_val_string = f"{round(self.temp_total_stream.values[-1]):3d}°C"
            lines0 = lines_temp[-1][: -len(last_val_string)] + last_val_string
            lines_temp = lines_temp[:-1] + [lines0]
            cpu_total_graph += "[color(5)]" + "\n".join(lines_temp) + "[/]"

        lines = [
            f"[{cpu_percent_colors[i]}]"
            + f"{self.cpu_percent_streams[i].graph[0]} "
            + f"{round(self.cpu_percent_streams[i].values[-1]):3d}%[/]"
            for i in self.core_order
        ]
        if self.has_temps:
            # add temperature in every other line
            for k, stream in enumerate(self.core_temp_streams):
                lines[
                    2 * k
                ] += f" [color(5)]{stream.graph[0]} {round(stream.values[-1])}°C[/]"

        # load_avg = os.getloadavg()
        # subtitle = f"Load Avg:  {load_avg[0]:.2f}  {load_avg[1]:.2f}  {load_avg[2]:.2f}"
        subtitle = f"{round(psutil.cpu_freq().current):4d} MHz"

        info_box = Panel(
            "\n".join(lines),
            title=self.box_title,
            title_align="left",
            subtitle=subtitle,
            subtitle_align="left",
            border_style="color(7)",
            box=box.SQUARE,
            expand=False,
        )

        # Manually adjust top margin. Waiting for vertical alignment in Rich.
        # <https://github.com/willmcgugan/rich/issues/1590>
        info_box_height = len(lines) + 2
        top_margin = (self.height - 2 - info_box_height) // 2
        info_box = Padding(info_box, (top_margin, 0, 0, 0))

        t = Table(expand=True, show_header=False, padding=0, box=None)
        # Add ratio 1 to expand that column as much as possible
        t.add_column("graph", no_wrap=True, ratio=1)
        t.add_column("box", no_wrap=True, justify="left")
        t.add_row(cpu_total_graph, info_box)

        self.panel = Panel(
            t,
            title=f"cpu - {self.brand_raw}",
            title_align="left",
            border_style="color(4)",
            box=box.SQUARE,
        )

        # textual method
        self.refresh()

    def render(self):
        if self.is_first_render:
            self.collect_data()
            self.is_first_render = False
        return self.panel

    async def on_resize(self, event):
        self.width = event.width
        self.height = event.height

        self.cpu_total_stream.reset_width(self.width - 35)

        if self.has_temps:
            # cpu total stream height: divide by two and round _down_
            self.cpu_total_stream.reset_height((self.height - 2) // 2)
            #
            self.temp_total_stream.reset_width(self.width - 35)
            # temp total stream height: divide by two and round _up_
            self.temp_total_stream.reset_height(-((2 - self.height) // 2))
        else:
            # full size cpu stream
            self.cpu_total_stream.reset_height(self.height - 2)



import getpass
import platform
import time
from datetime import datetime, timedelta

import distro
import psutil
from rich.table import Table
from textual.widget import Widget


class InfoLine(Widget):
    def on_mount(self):
        self.width = 0
        self.height = 0
        self.set_interval(1.0, self.refresh)

        # The getlogin docs say:
        # > For most purposes, it is more useful to use getpass.getuser() [...]
        # username = os.getlogin()
        username = getpass.getuser()
        ustring = f"{username} @"
        node = platform.node()
        if node:
            ustring += f" [b]{platform.node()}[/]"

        system = platform.system()
        if system == "Linux":
            ri = distro.os_release_info()
            system_list = [ri["name"]]
            if "version_id" in ri:
                system_list.append(ri["version_id"])
            system_list.append(f"{platform.architecture()[0]} / {platform.release()}")
            system_string = " ".join(system_list)
        elif system == "Darwin":
            system_string = f"macOS {platform.mac_ver()[0]}"
        else:
            # fallback
            system_string = ""

        self.left_string = " ".join([ustring, system_string])
        self.boot_time = psutil.boot_time()

    def render(self):
        uptime_s = time.time() - self.boot_time
        d = datetime(1, 1, 1) + timedelta(seconds=uptime_s)
        right = [f"up {d.day - 1}d, {d.hour}:{d.minute:02d}h"]

        battery = psutil.sensors_battery()
        if battery is not None:
            bat_style = "[color(1) reverse bold]" if battery.percent < 15 else ""
            bat_style_close = "[/]" if battery.percent < 15 else ""
            bat_symbol = "🔌" if battery.power_plugged else "🔋"
            right.append(
                f"{bat_style}{bat_symbol} {round(battery.percent)}%{bat_style_close}"
            )

        table = Table(show_header=False, expand=True, box=None, padding=0)
        if self.width < 100:
            table.add_column(justify="left", no_wrap=True)
            table.add_column(justify="right", no_wrap=True)
            table.add_row(self.left_string, ", ".join(right))
        else:
            table.add_column(justify="left", no_wrap=True, ratio=1)
            table.add_column(justify="center", no_wrap=True, ratio=1)
            table.add_column(justify="right", no_wrap=True, ratio=1)
            table.add_row(
                self.left_string, datetime.now().strftime("%c"), ", ".join(right)
            )
        return table

    async def on_resize(self, event):
        self.width = event.width
        self.height = event.height



from __future__ import annotations

from math import ceil

# String lookup and list lookup are equally fast, see
# <https://gist.github.com/nschloe/d790a873081dc504193c99d3758755b4>
num_to_braille = [
    [" ", "⢀", "⢠", "⢰", "⢸"],
    ["⡀", "⣀", "⣠", "⣰", "⣸"],
    ["⡄", "⣄", "⣤", "⣴", "⣼"],
    ["⡆", "⣆", "⣦", "⣶", "⣾"],
    ["⡇", "⣇", "⣧", "⣷", "⣿"],
]

num_to_braille_upside_down = [
    [" ", "⠈", "⠘", "⠸", "⢸"],
    ["⠁", "⠉", "⠙", "⠹", "⢹"],
    ["⠃", "⠋", "⠛", "⠻", "⢻"],
    ["⠇", "⠏", "⠟", "⠿", "⢿"],
    ["⡇", "⡏", "⡟", "⡿", "⣿"],
]


def _transpose(l):
    # https://stackoverflow.com/a/6473724/353337
    return map(list, zip(*l))


class BrailleStream:
    def __init__(
        self,
        width: int,
        height: int,
        minval: float,
        maxval: float,
        flipud: bool = False,
    ):
        # Store two alternating sets of graphs. This is necessary because
        # Braille symbols fit two values in each character.
        self._graphs = [
            [" " * width] * height,
            [" " * width] * height,
        ]
        self.graph_0_is_active = True
        self.width = width
        self.height = height
        self.minval = minval
        self.maxval = maxval
        self._last_blocks = [0] * height
        # store all values for resize purposes
        # we store one more value than what is displayed to account for the "old" graph
        self.values: list[float] = [minval] * (2 * self.width + 1)
        self.flipud = flipud
        self.lookup = num_to_braille_upside_down if flipud else num_to_braille

    def value_to_blocks(self, value: float):
        # value -> number of dots
        if value < self.minval:
            k = 0
        elif value > self.maxval:
            k = 4 * self.height
        else:
            k = ceil(
                (value - self.minval) / (self.maxval - self.minval) * 4 * self.height
            )
        # form blocks of 4
        blocks = [4] * (k // 4)
        if k % 4 > 0:
            blocks += [k % 4]
        blocks += [0] * (self.height - len(blocks))
        return blocks

    def add_value(self, value: float):
        blocks = self.value_to_blocks(value)

        chars = [self.lookup[i0][i1] for i0, i1 in zip(self._last_blocks, blocks)]
        if not self.flipud:
            chars = chars[::-1]

        # alternate between graphs[0] and graphs[1]
        self.graph_0_is_active = not self.graph_0_is_active
        g = self._graphs[0 if self.graph_0_is_active else 1]

        # update stream
        for k, char in enumerate(chars):
            g[k] = g[k][1:] + char

        self.values = self.values[1:] + [value]
        self._last_blocks = blocks

    @property
    def graph(self):
        return self._graphs[0 if self.graph_0_is_active else 1]

    def reset_width(self, width: int):
        if width == self.width:
            return
        elif width > self.width:
            diff = width - self.width
            self._graphs = [[" " * diff + row for row in g] for g in self._graphs]
            self.values = [self.minval] * (2 * diff) + self.values
        elif width < self.width:
            self._graphs = [[row[-width:] for row in g] for g in self._graphs]
            self.values = self.values[-(2 * width + 1) :]

        self.width = width

    def reset_height(self, height: int):
        if height == self.height:
            return

        # recreate both _graphs
        self.height = height
        blocks = [self.value_to_blocks(value) for value in self.values]

        assert len(self.values) == 2 * self.width + 1
        g = [
            # 0 -> 2 * k + 1
            [
                [
                    self.lookup[i0][i1]
                    for i0, i1 in zip(blocks[2 * k], blocks[2 * k + 1])
                ]
                for k in range(self.width)
            ],
            # 1 -> 2 * k + 2
            [
                [
                    self.lookup[i0][i1]
                    for i0, i1 in zip(blocks[2 * k + 1], blocks[2 * k + 2])
                ]
                for k in range(self.width)
            ],
        ]
        if not self.flipud:
            g = [
                [row[::-1] for row in g[0]],
                [row[::-1] for row in g[1]],
            ]
        # transpose and join
        self._graphs = [
            ["".join(row) for row in _transpose(g[0])],
            ["".join(row) for row in _transpose(g[1])],
        ]
        if self.graph_0_is_active:
            self._graphs = [self._graphs[1], self._graphs[0]]

        self._last_blocks = blocks[-1]



from textual.app import App

from ._cpu import CPU
from ._info import InfoLine
from ._mem import Mem
from ._net import Net
from ._procs_list import ProcsList


def run():
    # class TiptopApp(App):
    #     async def on_mount(self) -> None:
    #         await self.view.dock(InfoLine(), edge="top", size=1, name="info")
    #         await self.view.dock(CPU(), edge="top", size=14, name="cpu")
    #         await self.view.dock(ProcsList(), edge="right", size=70, name="proc")
    #         await self.view.dock(Mem(), edge="top", size=20, name="mem")
    #         await self.view.dock(Net(), edge="bottom", name="net")
    #
    #     async def on_load(self, _):
    #         await self.bind("i", "view.toggle('info')", "Toggle info")
    #         await self.bind("c", "view.toggle('cpu')", "Toggle cpu")
    #         await self.bind("m", "view.toggle('mem')", "Toggle mem")
    #         await self.bind("n", "view.toggle('net')", "Toggle net")
    #         await self.bind("p", "view.toggle('proc')", "Toggle proc")
    #         await self.bind("q", "quit", "quit")

    # with a grid
    class TiptopApp(App):
        async def on_mount(self) -> None:
            grid = await self.view.dock_grid(edge="left")

            # 34/55: approx golden ratio. See
            # <https://gist.github.com/nschloe/ab6c3c90b4a6bc02c40405803fa8fa35>
            # for the error.
            grid.add_column(fraction=34, name="left")
            grid.add_column(fraction=55, name="right")

            grid.add_row(size=1, name="topline")
            grid.add_row(fraction=1, name="top")
            grid.add_row(fraction=1, name="center")
            grid.add_row(fraction=1, name="bottom")

            grid.add_areas(
                area0="left-start|right-end,topline",
                area1="left-start|right-end,top",
                area2="left,center",
                area3="left,bottom",
                area4="right,center-start|bottom-end",
            )

            grid.place(
                area0=InfoLine(),
                area1=CPU(),
                area2=Mem(),
                area3=Net(),
                area4=ProcsList(),
            )

        async def on_load(self, _):
            await self.bind("q", "quit", "quit")

    TiptopApp.run()



# https://stackoverflow.com/a/1094933/353337
def sizeof_fmt(num, suffix: str = "iB", sep=" ", fmt=".0f"):
    assert num >= 0
    for unit in ["B", "K", "M", "G", "T", "P", "E", "Z"]:
        # actually 1024, but be economical with the return string size:
        if unit != "B":
            unit += suffix

        if num < 1000:
            string = f"{{:{fmt}}}".format(num)
            return f"{string}{sep}{unit}"
        num /= 1024
    string = f"{{:{fmt}}}".format(num)
    return f"{string}{sep}Y{suffix}"



from math import ceil

num_to_blockchar = [" ", "▁", "▂", "▃", "▄", "▅", "▆", "▇", "█"]


class BlockCharStream:
    def __init__(self, width: int, height: int, minval: float, maxval: float):
        self.graph = [" " * width] * height
        self.height = height
        self.minval = minval
        self.maxval = maxval
        self.last_value: float = minval

    def add_value(self, value):
        k = ceil((value - self.minval) / (self.maxval - self.minval) * 8 * self.height)

        blocks = [8] * (k // 8)
        if k % 8 > 0:
            blocks += [k % 8]
        blocks += [0] * (self.height - len(blocks))

        chars = [num_to_blockchar[i] for i in blocks]

        # update stream
        for k, char in enumerate(chars[::-1]):
            self.graph[k] = self.graph[k][1:] + char

        self.last_value = value
